#include <loader.h>
#include <s_globals.h>
#include <s_util.h>
#include <cleanup.h>
#include <matcher.h>
#include <rules.h>

p_plugin_list	ld_list=NULL;
mlock_t		ld_lock;

int loader_init(void)
{
	MUTEX_CREATE(ld_lock, "PluginList");

	cleanup_add_handler(cleanup_lock, (void *)&ld_lock);
	
	return 1;
}

void loader_load_plugin(char *path)
{
	p_plugin_list	plugin=NULL;
	plugin_in_t	export;
	plugin_out_t	import;
	
	if ( globals.num_plugins >= MAX_PLUGINS ) {
		print_out(PRINT_WARNING "MAX_PLUGINS reached (%i)" CRLF, 
			MAX_PLUGINS);
		return;
	}

	if ( path == NULL ) return;

	/* Allocate the mem for this plugin in the list */	
	if ( !(plugin=malloc(sizeof(plugin_list_t))) ) {
		return;
	}else{
		memset(plugin, 0, sizeof(plugin_list_t));
	}

	/* Open the plugin file */
	if ( !(plugin->lib=dlopen(path, PLUGIN_MODE )) ) {
		print_out(PRINT_ERROR "Couldn't open plugin (%s)" CRLF,path);
		free(plugin);
		return;
	}

	/* Resolve the API symbols */
	plugin->init   = dlsym(plugin->lib, PLUGIN_INIT);
	plugin->unload = dlsym(plugin->lib, PLUGIN_UNLOAD);

	if ( (plugin->init==NULL) || (plugin->unload==NULL) ) {
		print_out(PRINT_ERROR "Couldn't resolve plugin symbols (%s)" CRLF,path);
		dlclose(plugin->lib);
		free(plugin);
		return;
	}
	
	/* Setup stuff to export to plugin */
	memset(&export, 0, sizeof(export));
	memset(&import, 0, sizeof(import));
	export.size=sizeof(export);
	export.print_out=print_out;
	export.print_raw=print_raw;
	export.add_captype=capdev_add_type;
	export.add_demux=demux_add_handler;
	export.add_matcher=matcher_add;
	export.add_target=target_add;
	export.add_rule=matcher_add_rule;
	export.add_criteria=matcher_add_criteria_to_rule;
	export.rule_cleanup=matcher_rule_cleanup;
	export.set_target=matcher_set_rule_target;
	export.add_parser=rules_add_parser;
	export.list_append=list_append;
	export.list_remove=list_remove;
	export.list_prepend=list_prepend;
	template_fill_struct(&export.templates);
	
	switch ( plugin->init(&export, &import) ) {
		case PLUGIN_ERR_OK:
			print_out("Loaded %s plugin:\t [%s]" CRLF, import.name, import.desc);
			break;
	
		case PLUGIN_ERR_NOMEM:
			print_out(PRINT_ERROR "Plugin init failed (%s): "
				"Out of memory" CRLF, path);
			dlclose(plugin->lib);
			free(plugin);
			return;

		case PLUGIN_ERR_PARAM:
			print_out(PRINT_ERROR "Plugin init failed (%s): "
				"Bad parameter" CRLF, path);
			dlclose(plugin->lib);
			free(plugin);
			return;

		case PLUGIN_ERR_SIZE:
			print_out(PRINT_ERROR "Plugin init failed (%s): "
				"Wrong version, please recompile" CRLF, path);
			dlclose(plugin->lib);
			free(plugin);
			return;
			
		case PLUGIN_ERR_FAIL:
		default:
			print_out(PRINT_ERROR "Plugin init failed (%s): " 
				"Unspecified reason" CRLF, path);
			dlclose(plugin->lib);
			free(plugin);
			return;
	}
	
	plugin->name=import.name;
	plugin->desc=import.desc;
	
	CRITICAL_START(ld_lock);
	ld_list=(p_plugin_list)list_append((void *)plugin, (void *)ld_list);
	CRITICAL_STOP(ld_lock);

	globals.num_plugins++;
}

/* Remember a trailing slash on the dir!  */
int loader_load_dir(char *plugin_dir)
{
	DIR *dir=NULL;
	struct dirent *file=NULL;
	struct stat sfile;
	char fullpath[1024]; /* Bigger than PATH_MAX anywhere */

	if ( !(dir=opendir(plugin_dir)) )
		return 0;

	while(dir) {
		if ((file=readdir(dir)) == NULL) break;

		memset(&fullpath, 0, 256);
		strncpy((char *)&fullpath, plugin_dir, 255 );
		strncat((char *)&fullpath, file->d_name, 253 );

		if ( stat(fullpath, &sfile) ) continue;
		if ( (sfile.st_mode & S_IFREG) ) {
			loader_load_plugin((char *)&fullpath);
		}		
		
	}closedir(dir);
}

void loader_cleanup(int code, void *priv)
{
	p_plugin_list	tmp=NULL;
	p_plugin_list   foo=NULL;
	
	print_raw("Unloading all plugins" CRLF);

	CRITICAL_START(ld_lock);
	for (tmp=ld_list; tmp; tmp=(p_plugin_list)tmp->l.next) {
		if ( tmp->unload() != PLUGIN_ERR_OK ) {
			print_out(PRINT_WARNING "%s plugin reported "
				"problems unloading" CRLF, tmp->name);
		}
		dlclose(tmp->lib);
	}

	for (tmp=ld_list; tmp;) {
		foo=tmp;
		tmp=(p_plugin_list)tmp->l.next;
		free(foo);
		foo=NULL;
	}
	CRITICAL_STOP(ld_lock);
}
